(* ::Package:: *)

(* ::Title::Closed:: *)
(*OverwriteTarget*)


(* ::Input::Initialization:: *)
Quiet@*DeleteFile/@{"file.txt","file2.txt","file.gif"};


(* ::Title::Closed:: *)
(*FITS*)


(* ::Text:: *)
(*- Author is Mikayel Egibyan (mikayele@wolfram.com)*)
(*- Created for Wolfram Technology Conference 2018*)


(* ::Subsection::Closed:: *)
(*Init*)


$results = {};


folder = "";
itt = 0;
fileCount = 0;
percent = 0;
files = FileNames["*", folder];
exportDir;
metaToShow = <||>;
dataToShow = "";


CreateFITSLookupApplication[] := PopupWindow[Button["FITSLookup"], FITSLookup[], WindowSize -> {1300,550}, WindowTitle -> "FITS Lookup Application"]


(* ::Subsection::Closed:: *)
(*Caching*)


createCacheKeyMetadata[fileName_String] :=
	Hash[
		{
			FITSMetadata,
			{
				fileName
			}
		}
	]

createCacheKeyData[fileName_String] :=
	Hash[
		{
			FITSData,
			{
				fileName
			}
		}
	]


(* ::Subsection::Closed:: *)
(*FITS download*)


startExport[fileName_] :=
	Block[
		{
			fileShort = First[StringSplit[fileName, "."]],
			meta, metaCacheKey,
			image, table
		},

		Quiet[
			If[!TrueQ[DirectoryQ[exportDir]],
				CreateDirectory[exportDir];
				If[!TrueQ[DirectoryQ[exportDir]],
					CreateDialog[{
						TextCell["Export directory: "],
						Row[{
							InputField[Dynamic[exportDir], String, Enabled -> False, FieldHint -> "Export Directory"],
							Row[{Spacer[10], FileNameSetter[Dynamic[exportDir], "Directory"]}]
						}],
						DefaultButton[DialogReturn[startExport[fileName]]]
					}];
					Return[];
				];
			];
		];

		metaCacheKey = createCacheKeyMetadata[fileName];
		meta = Internal`CheckImageCache[metaCacheKey];
		Quiet@Export[FileNameJoin[{exportDir, fileShort <> "_MetaInformation.xml"}], meta];

		image = Import[fileName, "Image"];
		If[!SameQ[image, <||>],
			image = DeleteCases[DeleteCases[image, _?MissingQ], _?FailureQ];
			If[!SameQ[image, <||>],
				Quiet@Export[FileNameJoin[{exportDir, fileShort <> "_Image_HDU-" <> ToString[#] <> ".tiff"}], image[#]] & /@ Keys[image];
			];
		];

		table = Import[fileName, "TableData"];
		If[!SameQ[table, <||>],
			table = DeleteCases[DeleteCases[table, _?MissingQ], _?FailureQ];
			If[!SameQ[table, <||>],
				Quiet@Export[FileNameJoin[{exportDir, fileShort <> "_Table_HDU-" <> ToString[#] <> ".txt"}], table[#]] & /@ Keys[table];
			];
		];
	]


downloadDialog[fileName_, dest_] :=
	Block[
		{

		},
		exportDir = dest;

		CreateDialog[{
			TextCell["Export directory: "],
			Row[{
				InputField[Dynamic[exportDir], String, Enabled -> True, FieldHint -> "Export Directory"],
				Row[{Spacer[10], FileNameSetter[Dynamic[exportDir], "Directory"]}]
			}],
			DefaultButton[DialogReturn[startExport[fileName]]],
			TextCell["The file will be downloaded as follows:\n1. Image data will be stored in a TIFF file\n2. Table data will be stored in a TXT file\n3. Meta-information will be stored in a XML file\n"]
		}
		];
	]


(* ::Subsection::Closed:: *)
(*FITS lookup*)


(* ::Subsubsection::Closed:: *)
(*Load*)


FITSLookupCol[] :=
	Column[
		{
			Row[{
				fitsSearch[],
				Spacer[20],
				Dynamic[createResults[$results[[All, 1]], $results[[All, 2]]]],
				Spacer[20],
				Dynamic[TabView[{"MetaInformation" -> createMetaViewer[metaToShow], "Data" -> createDataViewer[dataToShow]}, ImageSize -> {450, 450}]]
			}]
		},
		Frame -> All,
		Background -> Nest[Lighter, LightGray, 3]
	]

FITSLookup[] =
	Column[
		{
			Row[
				{
					Spacer[500],
					Style[
						Text["FITS LOOKUP APPLICATION"],
						Nest[Darker, Blue, 2],
						22
					]
				}
			],
			FITSLookupCol[]
		},
		Frame -> All,
		Background -> Nest[Lighter, Black, 2]
	]


(* ::Subsubsection::Closed:: *)
(*Main*)


initializeDirectory[sourceDirectory_] :=
	findFits[
		sourceDirectory
		,
		"",
		"",
		"",
		Null
		,
		{Null, Null, Null},
		{Null, Null, Null, Null}
	]

findFits[
	sourceDirectory_
	,
	objName_,
	org_,
	dev_,
	year_
	,
	{lat_, lon_, lErr_},
	{bPos_, bMaj_, bMin_, bErr_}
] :=
	Catch[Block[
		{
			fitsFiles,
			cacheKey, cache,
			data, dataKey, dataCache,
			metadata,
			result = {},
			errL, errB,
			dates
		},
		itt = 0;
		fileCount = 0;
		$results = {};

		errL = If[NumberQ[lErr], lErr, 1];
		errB = If[NumberQ[bErr], bErr, 1];

		If[!StringQ[sourceDirectory] || !TrueQ[Quiet[DirectoryQ[sourceDirectory]]],
			Throw[Failure["InvDir", sourceDirectory <> " is not a valid directory"]];
		];

		SetDirectory[sourceDirectory];

		fitsFiles = FileNames["*", sourceDirectory];
		If[SameQ[fitsFiles, {}],
			Throw[Failure["ImprtFail", sourceDirectory <> " doesn't contain any file"]];
		];

		fitsFiles = FileNameTake[#, -1] & /@ fitsFiles;
		fitsFiles = Select[fitsFiles, SameQ[FileFormat[#], "FITS"] &];
		fileCount = Length[fitsFiles];

		Map[
			Function[singleFile,
				itt ++;
				cacheKey = createCacheKeyMetadata[singleFile];
				cache = Internal`CheckImageCache[cacheKey];
				If[FailureQ[cache],
					metadata =
						If[SameQ[FileFormat[singleFile], "FITS"],
							Quiet[Import[singleFile, "MetaInformation"]]
							,
							$Failed
						];
					Internal`SetImageCache[cacheKey, metadata];
					,
					metadata = cache;
				];

				If[!FailureQ[metadata],
					metadata = Join @@ Values[#] & /@ metadata;
					If[NumberQ[year],
						dates = DeleteCases[Map[Function[header, DeleteMissing[First /@ Values[KeySelect[header, StringContainsQ[#, "date", IgnoreCase -> True] &]]]], metadata], {}];
					];

					If[
						And[
							Or[SameQ[StringTrim[objName], ""],
								StringLength[objName] > 0 && AnyTrue[(StringQ[#] && StringContainsQ[#, ___ ~~ objName ~~ ___, IgnoreCase -> True]) & /@ First /@ Values[DeleteMissing[Lookup[#, "OBJECT"] & /@ metadata]], TrueQ]
							],
							Or[SameQ[StringTrim[org], ""],
								StringLength[org] > 0 && AnyTrue[(StringQ[#] && StringContainsQ[#, ___ ~~ org ~~ ___, IgnoreCase -> True]) & /@ First /@ Values[DeleteMissing[Lookup[#, "ORIGIN"] & /@ metadata]], TrueQ]
							],
							Or[SameQ[StringTrim[dev], ""],
								StringLength[dev] > 0 && AnyTrue[(StringQ[#] && StringContainsQ[#, ___ ~~ dev ~~ ___, IgnoreCase -> True]) & /@ First /@ Values[DeleteMissing[Lookup[#, "INSTRUME"] & /@ metadata]], TrueQ]
							],
							Or[!NumberQ[lat],
								AnyTrue[(QuantityQ[#] && Abs[Abs[QuantityMagnitude[#]] - Abs[lat]] < errL) & /@ First /@ Values[DeleteMissing[Lookup[#, "LATPOLE"] & /@ metadata]], TrueQ]
							],
							Or[!NumberQ[lon],
								AnyTrue[(QuantityQ[#] && Abs[Abs[QuantityMagnitude[#]] - Abs[lon]] < errL) & /@ First /@ Values[DeleteMissing[Lookup[#, "LONPOLE"] & /@ metadata]], TrueQ]
							],
							Or[!NumberQ[bMin],
								AnyTrue[(QuantityQ[#] && Abs[Abs[QuantityMagnitude[#]] - Abs[bMin]] < errB) & /@ First /@ Values[DeleteMissing[Lookup[#, "BMIN"] & /@ metadata]], TrueQ]
							],
							Or[!NumberQ[bMaj],
								AnyTrue[(QuantityQ[#] && Abs[Abs[QuantityMagnitude[#]] - Abs[bMaj]] < errB) & /@ First /@ Values[DeleteMissing[Lookup[#, "BMAJ"] & /@ metadata]], TrueQ]
							],
							Or[!NumberQ[bPos],
								AnyTrue[(QuantityQ[#] && Abs[Abs[QuantityMagnitude[#]] - Abs[bPos]] < errB) & /@ First /@ Values[DeleteMissing[Lookup[#, "BPA"] & /@ metadata]], TrueQ]
							],
							Or[!NumberQ[year],
								!SameQ[dates, <||>] &&
									AnyTrue[Flatten@Values@Map[
										Function[header,
											If[DateObjectQ[#] || TimeObjectQ[#],
												StringContainsQ[DateString[#], ToString[year]]
												,
												StringContainsQ[ToString[#], ToString[year]]
											] & /@ header
										],
										dates
									], TrueQ]
							]
						],

						dataKey = createCacheKeyData[singleFile];
						dataCache = Internal`CheckImageCache[dataKey];
						If[FailureQ[dataCache],
							data = Import[singleFile];
							Internal`SetImageCache[dataKey, data];
							,
							data = dataCache;
						];

						AppendTo[result,
							{
								Button[singleFile, (metaToShow = Internal`CheckImageCache[createCacheKeyMetadata[singleFile]]; dataToShow = Internal`CheckImageCache[createCacheKeyData[singleFile]]), ImageSize -> {Automatic, 25}],
								Button[
									Image[
										CompressedData["
														1:eJzNklsrhEEYxzdcuHThC/gGLl05JYeIkAtc7WZtSquWkvJZdreUkq/gihvH
														WESSG+ctkmPWzszj+c+B2bewNhem/vs+M8///9uZ952ayETfaFkoFJqs5J++
														8HRTIhGe6a/iyUB8ciwWj450xKeisWiiLlLOi9VWFSz6YiirR0GUOJBaqN36
														b4e0oZ17RU2rQgu13yuFt/ugqGNdaKH+C147s9r/CQ9eoX7m+Z5ihviGJ4rg
														OMvyraKlG/WxlvF4qJ0PHnj9rH9GjJNnczcaWekzqdcy9x7P3hf0Gu0dQsZn
														OD5090YU2zPZ5jVBi1dK+9t4DqHGGnrwwIuMf8clFb7b61eiyK6gVvZ3bQia
														PZLUsym0UGMNPXjgdXvTjOC57TPLvrBldnK+e9Oo07LQy74WZtw4flK0cKno
														RZi5+3bZHDMz5py9W0aosYae70UWDLDQr13Bu5IfHnd2vU/ut6wZaVbgjBjI
														goH+4LZ598lTWfCfPnN8X2r5LN+LLBhgDbMa+Lun+A6gn5PGB+W9nMvm1efc
														eZEFA6whVj3X8xfBN1v8QBaMIcvDN5s6lJQ+l5Tk/0oHlLIKrmsvZ5AFAyz/
														HmDPpci/V3PMx9PttVQ51jvqVsRz
														"
										],
										"Byte",
										ColorSpace -> "RGB",
										Interleaving -> True
									],
									downloadDialog[singleFile, FileNameJoin[{sourceDirectory, First@StringSplit[singleFile, "."]}]],
									ImageSize -> {Automatic, 25},
									Background -> GrayLevel[0.93],
									ContentPadding -> False,
									Method -> "Queued"
								]
							}
						];
					];
				];
			],
			fitsFiles
		];

		$results = result;
	]];


(* ::Subsection::Closed:: *)
(*GUI*)


(* ::Subsubsection::Closed:: *)
(*Search*)


fitsSearch[] :=
	DynamicModule[
		{
			sourceDirectory = folder
			,
			objName = "",
			org = "",
			dev = "",
			year
			,
			lat, lon, lErr,
			bMaj, bMin, bPos, bErr
			,
			controlsCreate,
			style,
			progress
		},

		progress = Row[{ProgressIndicator[Dynamic[If[Dynamic[fileCount] == 0, 0, Dynamic[itt] / Dynamic[fileCount]]], Background -> LightGray, ImageSize -> {170, Automatic}], Spacer[10], Dynamic[itt], "/", Dynamic[fileCount]}];
		style = {"Text"};

		controlsCreate =
			Pane[
				Grid[
					{
						{
							"Source directory:",
							Spacer[10],
							Column[{
								InputField[Dynamic[sourceDirectory], String, Enabled -> False, FieldHint -> "Source Directory"],
								Button[
									"Browse",
									sourceDirectory = SystemDialogInput["Directory", ""];
									If[StringQ[sourceDirectory],
										initializeDirectory[sourceDirectory]
									],
									Method -> "Queued"
								]
							}]
						}
						,
						{
							"Initialization progress:",
							Spacer[10],
							progress
						}
						,
						{
							Column[{"", ""}]
						}
						,
						{
							"Object name:",
							Spacer[10],
							InputField[Dynamic[objName], String, FieldSize -> 13, FieldHint -> "e.g. SNG - CAT"]
						}
						,
						{
							"Organization:",
							Spacer[10],
							InputField[Dynamic[org], String, FieldSize -> 13, FieldHint -> "e.g. ESO"]
						}
						,
						{
							"Device:",
							Spacer[10],
							InputField[Dynamic[dev], String, FieldSize -> 13, FieldHint -> "e.g. EGRET"]
						}
						,
						{
							"Year:",
							Spacer[10],
							InputField[Dynamic[year], Number, FieldSize -> 13, FieldHint -> "e.g. 2018"]
						}
						,
						{
							""
						}
						,
						{
							"Calestial coordinates:",
							Spacer[10],
							Row[
								{
									InputField[Dynamic[lat], Number, FieldSize -> 3, FieldHint -> "lat"],
									Spacer[3], ",", Spacer[3],
									InputField[Dynamic[lon], Number, FieldSize -> 3, FieldHint -> "lon"],
									Spacer[3], ",", Spacer[3],
									InputField[Dynamic[lErr], Number, FieldSize -> 3, FieldHint -> "err"]
								}
							]
						}
						,
						{
							"Beam:",
							Spacer[10],
							Row[
								{
									InputField[Dynamic[bPos], Number, FieldSize -> 3, FieldHint -> "pos"],
									Spacer[3], ",", Spacer[3],
									InputField[Dynamic[bMaj], Number, FieldSize -> 3, FieldHint -> "maj"],
									Spacer[3], ",", Spacer[3],
									InputField[Dynamic[bMin], Number, FieldSize -> 3, FieldHint -> "min"],
									Spacer[3], ",", Spacer[3],
									InputField[Dynamic[bErr], Number, FieldSize -> 3, FieldHint -> "err"]
								}
							]
						}
						,
						{
							""
						}
						,
						{
							Row[{
								Spacer[200],
								Button["Search",
									findFits[
										sourceDirectory
										,
										objName,
										org,
										dev,
										year
										,
										{lat, lon, lErr},
										{bPos, bMaj, bMin, bErr}
									],
									Method -> "Queued"
								],
								Spacer[10]
							}],
							SpanFromLeft
						}
						,
						{
							Column[{"", "", ""}]
						}
					}
					,
					Alignment -> Left,
					BaseStyle -> style,
					Background -> LightBlue
				]
				,
				{400, 460},
				Alignment -> {Right, Top},
				AppearanceElements -> None,
				ImageMargins -> 0
			]
	]



(* ::Subsubsection::Closed:: *)
(*Results*)


createResults[keys_, values_] :=
	DynamicModule[
		{
			scroller = 1, scrollx = 0, scrolly = 0, options
			,
			valuesM = List /@ values, keysM = List /@ keys
			,
			keysAllignment = {Center, Center}, keysDividers = {{1 -> White, -1 -> LightGray}, {1 -> True, -1 -> True}},
			keysItemSize = {28, 1.75}, keysSpacing = {2, 0.5}
			,
			valsAllignment = {Center, Center}, valsDividers = {{-1 -> White}, None},
			valsItemSize = {5, 1.75}, valsSpacing = {{2, {0.5}, 2}, 0.6}
			,
			keysOptions, valsOptions,
			w1 = 260, w2 = 70, h = 433, margin = 3
		},


		options = {
			Background -> {None, {{White, GrayLevel[0.93]}}},
			BaseStyle -> Directive[FontFamily -> "Helvetica", 11],
			FrameStyle -> Directive[Thin, GrayLevel[0.75]]
		};

		keysOptions = Join[options, {Alignment -> keysAllignment, ItemSize -> keysItemSize, Spacings -> keysSpacing}];
		valsOptions = Join[options, {Alignment -> valsAllignment, ItemSize -> valsItemSize, Spacings -> valsSpacing}];

		Dynamic[
			Grid[
				{
					{
						Pane[
							Grid[
								{
									{
										Style[Text["File Name"], Blue]
									}
								},
								keysOptions
							],
							{w1, All},
							Alignment -> {Left, Top},
							ImageMargins -> margin
						]
						,
						Pane[
							Grid[
								{
									{
										Style[Text["Save"], Blue]
									}
								},
								valsOptions
							],
							{w2, All},
							Alignment -> {Left, Top},
							ImageMargins -> margin
						]
					}
					,
					{
						Pane[
							Grid[
								keysM,
								keysOptions
							]
							,
							{w1, h},
							Alignment -> {Left, Top},
							AppearanceElements -> None,
							ImageMargins -> margin,
							ScrollPosition -> Dynamic[{scrollx, scrolly}]
						]
						,
						Pane[
							Grid[
								valuesM,
								valsOptions
							]
							,
							{w2, h},
							Alignment -> {Left, Top},
							AppearanceElements -> None,
							ImageMargins -> margin,
							ScrollPosition -> Dynamic[{scrollx, scrolly}],
							Scrollbars -> {False, True}
						]
					}
				}
				,
				Alignment -> {{Right, Left}, Top},
				Spacings -> {0, 0}
			]
		]
	];


(* ::Subsubsection::Closed:: *)
(*MetaInformation*)


toKeyVals[asc_] := {Keys[#], Values[#]} & @ (First /@ KeySort@KeyDrop[Join @@ Values[asc], {"HISTORY", "COMMENT"}])

createMetaViewer[meta_] :=
	Block[
		{
			keys, vals,
			res = {}
		},

		res =
			(
				"HDU " <> ToString[#] ->
					(
						{keys, vals} = toKeyVals[meta[#]];
						createMetaViewer[keys, vals]
					)
			) & /@ Keys[meta];

		If[res === {},
			createMetaViewer[{}, {}]
			,
			TabView[res]
		]
	]

createMetaViewer[keys_, values_] :=
	DynamicModule[
		{
			scroller = 1, scrollx = 0, scrolly = 0, options
			,
			valuesM = List /@ values, keysM = List /@ keys
			,
			keysAllignment = {Center, Center}, keysDividers = {{1 -> White, -1 -> LightGray}, {1 -> True, -1 -> True}},
			keysItemSize = {7, 1.75}, keysSpacing = {2, .5}
			,
			valsAllignment = {Center, Center}, valsDividers = {{-1 -> White}, None},
			valsItemSize = {30, 1.75}, valsSpacing = {{2, {.5}, 2}, .5}
			,
			keysOptions, valsOptions,
			w1 = 70, w2 = 320, h = 420, margin = 3
		},


		options = {
			Background -> {None, {{White, GrayLevel[0.93]}}},
			BaseStyle -> Directive[FontFamily -> "Helvetica", 11],
			FrameStyle -> Directive[Thin, GrayLevel[0.75]]
		};

		keysOptions = Join[options, {Alignment -> keysAllignment, ItemSize -> keysItemSize, Spacings -> keysSpacing}];
		valsOptions = Join[options, {Alignment -> valsAllignment, ItemSize -> valsItemSize, Spacings -> valsSpacing}];

		Dynamic[
			Grid[
				{
					{
						Pane[
							Grid[
								{
									{
										Style[Text["Key"], Blue, 14]
									}
								},
								keysOptions
							],
							{w1, All},
							Alignment -> {Left, Top},
							ImageMargins -> margin
						]
						,
						Pane[
							Grid[
								{
									{
										Style[Text["Value"], Blue, 14]
									}
								},
								valsOptions
							],
							{w2, All},
							Alignment -> {Left, Top},
							ImageMargins -> margin
						]
					}
					,
					{
						Pane[
							Grid[
								keysM,
								keysOptions
							]
							,
							{w1, h},
							Alignment -> {Left, Top},
							AppearanceElements -> None,
							ImageMargins -> margin,
							ScrollPosition -> Dynamic[{scrollx, scrolly}]
						]
						,
						Pane[
							Grid[
								valuesM,
								valsOptions
							]
							,
							{w2, h},
							Alignment -> {Left, Top},
							AppearanceElements -> None,
							ImageMargins -> margin,
							ScrollPosition -> Dynamic[{scrollx, scrolly}],
							Scrollbars -> {True, True}
						]
					}
				}
				,
				Alignment -> {{Right, Left}, Top},
				Spacings -> {0, 0}
			]
		]
	];



(* ::Subsubsection::Closed:: *)
(*Data*)


createDataViewer[data_Association] :=
	Block[
		{
			keys = Keys[data],
			res = {}
		},

		res =
			(
				"HDU " <> ToString[#] ->
					(
						createDataViewer[data[#]]
					)
			) & /@ Keys[data];

		If[res === {},
			createDataViewer[""]
			,
			TabView[res]
		]
	]

createDataViewer[data_] :=
	DynamicModule[
		{
			scroller = 1, scrollx = 0, scrolly = 0, options
			,
			keysAllignment = {Center, Center}, keysDividers = {{1 -> White, -1 -> LightGray}, {1 -> True, -1 -> True}},
			keysItemSize = {7, 1.75}, keysSpacing = {2, .5}
			,
			keysOptions,
			w = 390, h = 420, margin = 3
		},


		options = {
			Background -> {None, {{White, GrayLevel[0.93]}}},
			BaseStyle -> Directive[FontFamily -> "Helvetica", 11],
			FrameStyle -> Directive[Thin, GrayLevel[0.75]]
		};

		keysOptions = Join[options, {Alignment -> keysAllignment, ItemSize -> keysItemSize, Spacings -> keysSpacing}];

		Dynamic[
			Grid[
				{
					{
						Pane[
							data
							,
							{w, h},
							Alignment -> {Left, Top},
							AppearanceElements -> None,
							ImageMargins -> margin,
							ScrollPosition -> Dynamic[{scrollx, scrolly}],
							Scrollbars -> {False, True}
						]
					}
				}
				,
				Alignment -> {{Right, Left}, Top},
				Spacings -> {0, 0}
			]
		]
	];
